#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "iscm0.h"

int yyparse();
static EXPS* exps = NULL;

int
main()
{
  EXPS *p;
  if(yyparse() == 0){
    p = exps;
    while (p){
      eval(p->pexp, NULL);
      p = p->next;
    }
  }
  return 0;
}

void*
mycalloc(int size)
{
  void* p;
  if((p = calloc(1, size)) == NULL){
    fprintf(stderr, "Fatal: malloc faild.\n");
    exit(1);
  }
  return p;
}

static OBJ*
new_obj()
{
  return mycalloc(sizeof(OBJ));
}

void
append_exp(OBJ *const pexp)
{
  EXPS *p, *node;

  node = mycalloc(sizeof(EXPS));
  node->pexp = pexp;

  if (exps){
    p = exps;
    while (p->next){
      p = p->next;
    }
    p->next = node;
  } else {
    exps = node;
  }
}

OBJ*
create_pair(OBJ *const p)
{
  OBJ* ret;
  ret = new_obj();
  ret->ty = PAIR_OBJ;
  ret->u.car = p;
  ret->cdr = NULL;
  return ret;
}

OBJ* 
append_to_list(OBJ *const plist, OBJ *const pobj)
{
  OBJ* p;
  p = plist;
  while(p->cdr){
    p = p->cdr;
  }
  p->cdr = create_pair(pobj);
  return plist;
}

OBJ*
nil()
{
  return NULL;
}

void* 
create_obj_int(long ival)
{
  OBJ* p;
  p = new_obj();
  p->ty = INT_OBJ;
  p->u.ival = ival;
  return p;
}

void* 
create_obj_str(const char *const str)
{
  OBJ* p;
  const char* ps;
  char *pd;

  p = new_obj();
  p->ty = STR_OBJ;
  p->u.str = mycalloc(strlen(str)); /* enough */

  ps = &str[1]; /* skip the first " */
  pd = p->u.str;

  while(*ps != '\0' && *ps != '"'){
    if(*ps == '\\' && *(ps + 1) == 'n'){
      *pd++ = '\n';
      ps += 2;
    } else {
      *pd++ = *ps++;
    }
  }

  *pd = '\0';

  return p;
}

void* 
create_obj_id(const char *const str)
{
  OBJ* p;

  p = new_obj();
  p->ty = ID_OBJ;
  p->u.str = mycalloc(strlen(str) + 1);
  strcpy(p->u.str, str);

  return p;
}

static void
print_obj_sub(const OBJ* obj)
{
  if(obj){
    switch(obj->ty){
    case INT_OBJ:  
      printf("%ld", obj->u.ival); 
      break;
    case STR_OBJ:  
      printf("\"%s\"", obj->u.str); 
      break;
    case ID_OBJ:   
      printf("%s", obj->u.str); 
      break;
    case PAIR_OBJ:
      printf("(");
      print_obj_sub(obj->u.car);
      printf(" . ");
      print_obj_sub(obj->cdr);
      printf(")");
      break;
    case FUNC:
      printf("#<closure>");
      break;
    case BUILTIN_FUNC:
      printf("#<built-in>");
    }
  }else{
    printf("#<nil>");
  }
}

void
print_obj(const OBJ* obj)
{
  print_obj_sub(obj);
  printf("\n");
}
